from ruamel.yaml.comments import CommentedSeq, CommentedMap
from strictyaml.exceptions import raise_type_error
from ruamel.yaml import RoundTripDumper
from ruamel.yaml import dump
from strictyaml import utils
from copy import deepcopy
import decimal


class YAML(object):
    def __init__(self, value, text=None, document=None, location=None):
        self._value = value
        self._text = str(value) if text is None else text
        self._document = deepcopy(document)
        self._location = location

    def __int__(self):
        return int(self._value)

    def __str__(self):
        if type(self._value) in (str, int, float, decimal.Decimal):
            return str(self._value)
        elif isinstance(self._value, CommentedMap) or isinstance(self._value, CommentedSeq):
            raise TypeError(
                "Cannot cast mapping/sequence '{0}' to string".format(repr(self._value))
            )
        else:
            raise_type_error(
                repr(self), "str", "str(yamlobj.value) or str(yamlobj.text)"
            )

    @property
    def data(self):
        if type(self._value) is CommentedMap:
            mapping = {}
            for key, value in self._value.items():
                if type(key) is str:
                    mapping[key] = value.data
                else:
                    mapping[key.data] = value.data
            return mapping
        elif type(self._value) is CommentedSeq:
            return [item.data for item in self._value]
        else:
            return self._value

    def as_marked_up(self):
        """
        Returns ruamel.yaml CommentedSeq/CommentedMap objects
        with comments. This can be fed directly into a ruamel.yaml
        dumper.
        """
        if isinstance(self._value, CommentedMap):
            new_commented_map = deepcopy(self._value)

            for key, value in self._value.items():
                del new_commented_map[key._value]
                new_commented_map[key._value] = value.as_marked_up()
            return new_commented_map
        elif isinstance(self._value, CommentedSeq):
            new_commented_seq = deepcopy(self._value)

            for i, item in enumerate(self._value):
                new_commented_seq[i] = item.as_marked_up()
            return new_commented_seq
        else:
            if utils.is_integer(self._text):
                return int(self._text)
            elif utils.is_decimal(self._text):
                return decimal.Decimal(self._text)
            else:
                return self._text

    @property
    def start_line(self):
        """
        Return line number that the element starts on (including preceding comments).
        """
        return self._location.start_line(self._document)

    @property
    def end_line(self):
        """
        Return line number that the element ends on (including trailing comments).
        """
        return self._location.end_line(self._document)

    def lines(self):
        """
        Return a string of the lines which make up the selected line
        including preceding and trailing comments.
        """
        return self._location.lines(self._document)

    def lines_before(self, how_many):
        return self._location.lines_before(self._document, how_many)

    def lines_after(self, how_many):
        return self._location.lines_after(self._document, how_many)

    def __float__(self):
        return float(self._value)

    def __repr__(self):
        return u"YAML({0})".format(self.data)

    def __bool__(self):
        if isinstance(self._value, bool):
            return self._value
        else:
            raise_type_error(
                repr(self), "bool", "bool(yamlobj.value) or bool(yamlobj.text)"
            )

    def __nonzero__(self):
        return self.__bool__()

    def __getitem__(self, index):
        return self._value[index]

    def __setitem__(self, index, value):
        if isinstance(value, YAML):
            self._value[index] = value.copy()
        else:
            self._value[index] = YAML(value)

    def __delitem__(self, index):
        del self._value[index]

    def __hash__(self):
        return hash(self._value)

    def __len__(self):
        return len(self._value)

    def as_yaml(self):
        return dump(self.as_marked_up(), Dumper=RoundTripDumper)

    def items(self):
        if not isinstance(self._value, CommentedMap):
            raise TypeError("{0} not a mapping, cannot use .items()".format(repr(self)))
        return [(key, self._value[key]) for key, value in self._value.items()]

    def keys(self):
        if not isinstance(self._value, CommentedMap):
            raise TypeError("{0} not a mapping, cannot use .keys()".format(repr(self)))
        return [key for key, value in self._value.items()]

    def values(self):
        if not isinstance(self._value, CommentedMap):
            raise TypeError("{0} not a mapping, cannot use .values()".format(repr(self)))
        return [self._value[key] for key, value in self._value.items()]

    def get(self, index, default=None):
        if not isinstance(self._value, CommentedMap):
            raise TypeError("{0} not a mapping, cannot use .get()".format(repr(self)))
        return self._value[index] if index in self._value.keys() else default

    def __contains__(self, item):
        if isinstance(self._value, CommentedSeq):
            return item in self._value
        elif isinstance(self._value, CommentedMap):
            return item in self.keys()
        else:
            return item in self._value

    @property
    def text(self):
        if isinstance(self._value, CommentedMap):
            raise TypeError("{0} is a mapping, has no text value.".format(repr(self)))
        if isinstance(self._value, CommentedSeq):
            raise TypeError("{0} is a sequence, has no text value.".format(repr(self)))
        return self._text

    def copy(self):
        return deepcopy(self)

    def __gt__(self, val):
        if isinstance(self._value, CommentedMap) or isinstance(self._value, CommentedSeq):
            raise TypeError("{0} not an orderable type.".format(repr(self._value)))
        return self._value > val

    def __lt__(self, val):
        if isinstance(self._value, CommentedMap) or isinstance(self._value, CommentedSeq):
            raise TypeError("{0} not an orderable type.".format(repr(self._value)))
        return self._value < val

    @property
    def value(self):
        return self._value

    def is_mapping(self):
        return isinstance(self._value, CommentedMap)

    def is_sequence(self):
        return isinstance(self._value, CommentedSeq)

    def is_scalar(self):
        return not isinstance(self._value, CommentedSeq) \
            and not isinstance(self._value, CommentedMap)

    def __eq__(self, value):
        return self.data == value



from ruamel.yaml.comments import CommentedSeq, CommentedMap
from strictyaml.exceptions import YAMLValidationError
from strictyaml.yamllocation import YAMLLocation
from strictyaml.exceptions import raise_exception
from strictyaml.representation import YAML
from strictyaml import utils
import dateutil.parser
import decimal
import copy


class Optional(object):
    def __init__(self, key):
        self.key = key


class Validator(object):
    def __or__(self, other):
        return OrValidator(self, other)

    def __call__(self, document, location=None):
        if location is None:
            location = YAMLLocation()
        return self.validate(document, location)


class OrValidator(Validator):
    def __init__(self, validator_a, validator_b):
        self._validator_a = validator_a
        self._validator_b = validator_b

    def validate(self, document, location=None):
        if location is None:
            location = YAMLLocation()
            document = copy.deepcopy(document)

        try:
            return self._validator_a(document, location=location)
        except YAMLValidationError:
            return self._validator_b(document, location=location)

    def __repr__(self):
        return u"{0} | {1}".format(repr(self._validator_a), repr(self._validator_b))


def schema_from_data(document):
    if isinstance(document, CommentedMap):
        return Map({key: schema_from_data(value) for key, value in document.items()})
    elif isinstance(document, CommentedSeq):
        return FixedSeq([schema_from_data(item) for item in document])
    else:
        return Str()


class Any(Validator):
    """
    Validates any YAML and returns simple dicts/lists of strings.
    """
    def validate(self, document, location=None):
        if location is None:
            location = YAMLLocation()
            document = copy.deepcopy(document)
        return schema_from_data(location.get(document))(document, location=location)

    def __repr__(self):
        return u"Any()"


class Scalar(Validator):
    @property
    def rule_description(self):
        return "a {0}".format(self.__class__.__name__.lower())

    def validate(self, document, location=None):
        val = location.get(document)

        if type(val) == CommentedSeq or type(val) == CommentedMap:
            raise_exception(
                "when expecting a {0}".format(self.__class__.__name__.lower()),
                "found mapping/sequence",
                document, location=location,
            )
        else:
            return self.validate_scalar(document, location, value=None)


class Enum(Scalar):
    def __init__(self, restricted_to):
        for element in restricted_to:
            assert type(element) is str
        self._restricted_to = restricted_to

    def validate_scalar(self, document, location, value):
        val = str(location.get(document)) if value is None else value
        if val not in self._restricted_to:
            raise_exception(
                "when expecting one of: {0}".format(", ".join(self._restricted_to)),
                "found '{0}'".format(val),
                document, location=location,
            )
        else:
            return YAML(val, document=document, location=location)

    def __repr__(self):
        return u"Enum({0})".format(repr(self._restricted_to))


class EmptyNone(Scalar):
    def validate_scalar(self, document, location, value):
        val = str(location.get(document)) if value is None else value
        if val != "":
            raise_exception(
                "when expecting an empty value",
                "found non-empty value",
                document, location=location,
            )
        else:
            return self.empty(document, location)

    def empty(self, document, location):
        return YAML(None, '', document=document, location=location)

    def __repr__(self):
        return u"EmptyNone()"


class EmptyDict(EmptyNone):
    def empty(self, document, location):
        return YAML({}, '', document=document, location=location)

    def __repr__(self):
        return u"EmptyDict()"


class EmptyList(EmptyNone):
    def empty(self, document, location):
        return YAML([], '', document=document, location=location)

    def __repr__(self):
        return u"EmptyList()"


class CommaSeparated(Scalar):
    def __init__(self, item_validator):
        self._item_validator = item_validator

    def validate_scalar(self, document, location, value):
        val = str(location.get(document)) if value is None else value
        return YAML(
            [
                YAML(self._item_validator.validate_scalar(document, location, value=item.lstrip()))
                for item in val.split(",")
            ],
            document=document,
            location=location
        )


class Str(Scalar):
    def validate_scalar(self, document, location, value=None):
        return YAML(
            str(location.get(document)) if value is None else value,
            text=location.get(document),
            document=document,
            location=location
        )

    def __repr__(self):
        return u"Str()"


class Int(Scalar):
    def validate_scalar(self, document, location, value=None):
        val = str(location.get(document)) if value is None else value
        if not utils.is_integer(val):
            raise_exception(
                    "when expecting an integer",
                    "found non-integer",
                    document, location=location,
                )
        else:
            return YAML(int(val), val, document=document, location=location)

    def __repr__(self):
        return u"Int()"


TRUE_VALUES = ["yes", "true", "on", "1", ]
FALSE_VALUES = ["no", "false", "off", "0", ]
BOOL_VALUES = TRUE_VALUES + FALSE_VALUES


class Bool(Scalar):
    def validate_scalar(self, document, location, value=None):
        val = str(location.get(document)) if value is None else value
        if str(val).lower() not in BOOL_VALUES:
            raise_exception(
                """when expecting a boolean value (one of "{0}")""".format(
                    '", "'.join(BOOL_VALUES)
                ),
                "found non-boolean",
                document, location=location,
            )
        else:
            if val in TRUE_VALUES:
                return YAML(True, val, document=document, location=location)
            else:
                return YAML(False, val, document=document, location=location)

    def __repr__(self):
        return u"Bool()"


class Float(Scalar):
    def validate_scalar(self, document, location, value=None):
        val = str(location.get(document)) if value is None else value
        if not utils.is_decimal(str(val)):
            raise_exception(
                "when expecting a float",
                "found non-float",
                document, location=location,
            )
        else:
            return YAML(float(val), val, document=document, location=location)

    def __repr__(self):
        return u"Float()"


class Decimal(Scalar):
    def validate_scalar(self, document, location, value=None):
        val = str(location.get(document)) if value is None else value
        if not utils.is_decimal(str(val)):
            raise_exception(
                "when expecting a decimal",
                "found non-decimal",
                document, location=location,
            )
        else:
            return YAML(decimal.Decimal(val), val, document=document, location=location)

    def __repr__(self):
        return u"Decimal()"


class Datetime(Scalar):
    def validate_scalar(self, document, location, value=None):
        val = str(location.get(document)) if value is None else value

        try:
            return YAML(dateutil.parser.parse(val), val, document=document, location=location)
        except ValueError:
            raise_exception(
                "when expecting a datetime",
                "found non-datetime",
                document, location=location,
            )

    def __repr__(self):
        return u"Datetime()"


class MapPattern(Validator):
    def __init__(self, key_validator, value_validator):
        self._key_validator = key_validator
        self._value_validator = value_validator

    def validate(self, document, location=None):
        if location is None:
            location = YAMLLocation()
            document = copy.deepcopy(document)
        return_snippet = location.get(document)

        if type(location.get(document)) != CommentedMap:
            raise_exception(
                "when expecting a mapping",
                "found non-mapping",
                document, location=location,
            )
        else:
            for key, value in location.get(document).items():
                valid_key = self._key_validator(document, location.key(key))
                valid_val = self._value_validator(document, location.val(key))
                return_snippet[valid_key] = valid_val

        return YAML(return_snippet, document=document, location=location)

    def __repr__(self):
        return u"MapPattern({0}, {1})".format(
            repr(self._key_validator), repr(self._value_validator)
        )


class Map(Validator):
    def __init__(self, validator):
        self._validator = validator

        self._validator_dict = {
            key.key if type(key) == Optional else key: value for key, value in validator.items()
        }

    def __repr__(self):
        return u"Map({{{0}}})".format(', '.join([
            '{0}: {1}'.format(
                'Optional("{0}")'.format(key.key) if type(key) is Optional else '"{0}"'.format(key),
                repr(value)
            ) for key, value in self._validator.items()
        ]))

    def validate(self, document, location=None):
        if location is None:
            location = YAMLLocation()
            document = copy.deepcopy(document)
        return_snippet = location.get(document)

        if type(location.get(document)) != CommentedMap:
            raise_exception(
                "when expecting a mapping",
                "found non-mapping",
                document, location=location,
            )
        else:
            for key, value in location.get(document).items():
                if key not in self._validator_dict.keys():
                    raise_exception(
                        "while parsing a mapping",
                        "unexpected key not in schema '{0}'".format(key),
                        document, location=location.key(key)
                    )

                del return_snippet[key]
                return_snippet[
                    YAML(key, document=document, location=location.key(key))
                ] = self._validator_dict[key](
                    document, location.val(key)
                )

        return YAML(return_snippet, document=document, location=location)


class Seq(Validator):
    def __init__(self, validator):
        self._validator = validator

    def __repr__(self):
        return "Seq({0})".format(repr(self._validator))

    def validate(self, document, location=None):
        if location is None:
            location = YAMLLocation()
            document = copy.deepcopy(document)
        return_snippet = location.get(document)

        if type(location.get(document)) != CommentedSeq:
            raise_exception(
                "when expecting a sequence",
                "found non-sequence",
                document, location=location,
            )
        else:
            for i, item in enumerate(location.get(document)):
                return_snippet[i] = self._validator(document, location=location.index(i))

        return YAML(return_snippet, document=document, location=location)


class FixedSeq(Validator):
    def __init__(self, validators):
        self._validators = validators

    def __repr__(self):
        return "FixedSeq({0})".format(repr(self._validators))

    def validate(self, document, location=None):
        if location is None:
            location = YAMLLocation()
            document = copy.deepcopy(document)
        return_snippet = location.get(document)

        if type(location.get(document)) != CommentedSeq:
            raise_exception(
                "when expecting a sequence of {0} elements".format(len(self._validators)),
                "found non-sequence",
                document, location=location,
            )
        else:
            if len(self._validators) != len(location.get(document)):
                raise_exception(
                    "when expecting a sequence of {0} elements".format(len(self._validators)),
                    "found a sequence of {0} elements".format(len(location.get(document))),
                    document, location=location,
                )
            for i, item_and_val in enumerate(zip(location.get(document), self._validators)):
                item, validator = item_and_val
                return_snippet[i] = validator(document, location=location.index(i))

        return YAML(return_snippet, document=document, location=location)


class UniqueSeq(Validator):
    def __init__(self, validator):
        self._validator = validator

    def __repr__(self):
        return "UniqueSeq({0})".format(repr(self._validator))

    def validate(self, document, location=None):
        if location is None:
            location = YAMLLocation()
            document = copy.deepcopy(document)

        return_snippet = location.get(document)

        if type(location.get(document)) != CommentedSeq:
            raise_exception(
                "when expecting a unique sequence",
                "found non-sequence",
                document, location=location,
            )
        else:
            existing_items = set()

            for i, item in enumerate(location.get(document)):
                if item in existing_items:
                    raise_exception(
                        "while parsing a sequence",
                        "duplicate found",
                        document, location=location
                    )
                else:
                    existing_items.add(item)
                    return_snippet[i] = self._validator(document, location=location.index(i))

        return YAML(return_snippet, document=document, location=location)



# The all important loader
from strictyaml.parser import load

# YAML object
from strictyaml.representation import YAML

# Validators
from strictyaml.validators import Optional
from strictyaml.validators import Validator
from strictyaml.validators import OrValidator
from strictyaml.validators import Any
from strictyaml.validators import Scalar
from strictyaml.validators import Enum
from strictyaml.validators import Str
from strictyaml.validators import Int
from strictyaml.validators import Bool
from strictyaml.validators import Float
from strictyaml.validators import Decimal
from strictyaml.validators import Map
from strictyaml.validators import MapPattern
from strictyaml.validators import Seq
from strictyaml.validators import UniqueSeq
from strictyaml.validators import FixedSeq
from strictyaml.validators import Datetime
from strictyaml.validators import EmptyNone
from strictyaml.validators import EmptyDict
from strictyaml.validators import EmptyList
from strictyaml.validators import CommaSeparated

# Base exception from ruamel.yaml (all exceptions inherit from this)
from ruamel.yaml import YAMLError

# Exceptions
from strictyaml.exceptions import StrictYAMLError
from strictyaml.exceptions import YAMLValidationError

# Disallowed token exceptions
from strictyaml.exceptions import DisallowedToken

from strictyaml.exceptions import TagTokenDisallowed
from strictyaml.exceptions import FlowMappingDisallowed
from strictyaml.exceptions import AnchorTokenDisallowed
from strictyaml.exceptions import DuplicateKeysDisallowed



from ruamel.yaml import MarkedYAMLError
from ruamel.yaml.dumper import RoundTripDumper
from ruamel.yaml import dump
try:
    from ruamel.yaml.error import Mark as StringMark
except ImportError:
    from ruamel.yaml.error import StringMark


class StrictYAMLError(MarkedYAMLError):
    pass


class YAMLValidationError(StrictYAMLError):
    pass


class DisallowedToken(StrictYAMLError):
    MESSAGE = "Disallowed token"


class TagTokenDisallowed(DisallowedToken):
    MESSAGE = "Tag tokens not allowed"


class FlowMappingDisallowed(DisallowedToken):
    MESSAGE = "Flow mapping tokens not allowed"


class AnchorTokenDisallowed(DisallowedToken):
    MESSAGE = "Anchor tokens not allowed"


class DuplicateKeysDisallowed(DisallowedToken):
    MESSAGE = "Duplicate keys not allowed"


def raise_exception(context, problem, document, location):
    str_document = dump(document, Dumper=RoundTripDumper)
    context_line = location.start_line(document) - 1
    problem_line = location.end_line(document) - 1
    context_index = len('\n'.join(str_document.split('\n')[:context_line]))
    problem_index = len('\n'.join(str_document.split('\n')[:problem_line]))
    raise YAMLValidationError(
        context,
        StringMark(
            "<unicode string>", context_index, context_line, 0, str_document, context_index + 1
        ),
        problem,
        StringMark(
            "<unicode string>", problem_index, problem_line, 0, str_document, problem_index + 1
        )
    )


def raise_type_error(yaml_object, to_type, alternatives):
    raise TypeError((
        "Cannot cast {0} to {1}.\n"
        "Use {2} instead."
    ).format(repr(yaml_object), to_type, alternatives))
